import logging
from django.utils import timezone
from django_keycloak_lma.models import ExchangedToken
import django_keycloak_lma.services.oidc_profile
import django_keycloak_lma.services.messages as messages
from django_keycloak_lma.services import exceptions

logger = logging.getLogger(__name__)


def exchange_remote_token(client, access_token):
    """
    Exchange remote client's access token for a token of OpenID Connect profile.

    :param client: current client
    :param access_token: remote access token
    :rtype: dict
    """
    issuer = django_keycloak_lma.services.realm.get_issuer(client.realm)
    # remote token decoding check
    id_token_object = client.openid_api_client.decode_token(
        token=access_token,
        key=client.realm.certs,
        algorithms=client.openid_api_client.well_known[
            'id_token_signing_alg_values_supported'],
        issuer=issuer
    )

    if not django_keycloak_lma.services.oidc_profile.check_groups_for_client_access(client.client_id, id_token_object['user_group']):
        raise exceptions.AuthenticationFailed(detail=messages.get_message(messages.ABSENCE_USER_GROUP_ERROR))

    oidc_profile = django_keycloak_lma.services.oidc_profile.get_from_id_token(client, access_token)

    try:
        # try exchange
        token_response = client.openid_api_client.token_exchange(
            subject_token=access_token,
            requested_token_type='urn:ietf:params:oauth:token-type:refresh_token'
        )
    except Exception:
        raise exceptions.AuthenticationFailed(detail=messages.get_message(messages.REMOTE_REFRESH_TOKEN_ERROR))

    initiate_time = timezone.now()

    oidc_profile = django_keycloak_lma.services.oidc_profile.update_tokens(token_model=oidc_profile,
                                                                           token_response=token_response,
                                                                           initiate_time=initiate_time)
    return oidc_profile


def exchange_token(oidc_profile, remote_client):
    """
    Exchange access token from OpenID Connect profile for a token of given
    remote client.

    :param django_keycloak_lma.models.OpenIdConnectProfile oidc_profile:
    :param django_keycloak_lma.models.RemoteClient remote_client:
    :rtype: dict
    """
    active_access_token = django_keycloak_lma.services.oidc_profile \
        .get_active_access_token(oidc_profile=oidc_profile)

    # http://www.keycloak.org/docs/latest/securing_apps/index.html#_token-exchange
    return oidc_profile.realm.client.openid_api_client.token_exchange(
        audience=remote_client.name,
        subject_token=active_access_token,
        requested_token_type='urn:ietf:params:oauth:token-type:refresh_token'
    )


def get_active_remote_client_token(oidc_profile, remote_client):
    """
    Get an active remote client token. Exchange when not available or expired.

    :param django_keycloak_lma.models.OpenIdConnectProfile oidc_profile:
    :param django_keycloak_lma.models.RemoteClient remote_client:
    :rtype: str
    """
    exchanged_token, _ = ExchangedToken.objects.get_or_create(
        oidc_profile=oidc_profile,
        remote_client=remote_client
    )

    initiate_time = timezone.now()

    if exchanged_token.refresh_expires_before is None \
            or initiate_time > exchanged_token.refresh_expires_before \
            or initiate_time > exchanged_token.expires_before:
        token_response = exchange_token(oidc_profile, remote_client)

        exchanged_token = django_keycloak_lma.services.oidc_profile.update_tokens(
            token_model=exchanged_token,
            token_response=token_response,
            initiate_time=initiate_time
        )

    return exchanged_token.access_token
